@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2015-2020 Canonical Ltd
+ * Copyright (C) 2015-2022 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -22,7 +22,10 @@ package asserts
 import (
 	"bytes"
 	"crypto"
+	"encoding/hex"
+	"errors"
 	"fmt"
+	"strings"
 	"time"
 
 	// expected for digests
@@ -31,6 +34,7 @@ import (
 	"github.com/snapcore/snapd/osutil"
 	"github.com/snapcore/snapd/release"
 	"github.com/snapcore/snapd/snap/naming"
+	"github.com/snapcore/snapd/strutil"
 )
 
 // SnapDeclaration holds a snap-declaration assertion, declaring a
@@ -38,12 +42,13 @@ import (
 // publisher and its other properties.
 type SnapDeclaration struct {
 	assertionBase
-	refreshControl []string
-	plugRules      map[string]*PlugRule
-	slotRules      map[string]*SlotRule
-	autoAliases    []string
-	aliases        map[string]string
-	timestamp      time.Time
+	refreshControl      []string
+	plugRules           map[string]*PlugRule
+	slotRules           map[string]*SlotRule
+	autoAliases         []string
+	aliases             map[string]string
+	revisionAuthorities []*RevisionAuthority
+	timestamp           time.Time
 }
 
 // Series returns the series for which the snap is being declared.
@@ -97,6 +102,21 @@ func (snapdcl *SnapDeclaration) Aliases() map[string]string {
 	return snapdcl.aliases
 }
 
+// RevisionAuthority return any revision authority entries matching the given
+// provenance.
+func (snapdcl *SnapDeclaration) RevisionAuthority(provenance string) []*RevisionAuthority {
+	res := make([]*RevisionAuthority, 0, 1)
+	for _, ra := range snapdcl.revisionAuthorities {
+		if strutil.ListContains(ra.Provenance, provenance) {
+			res = append(res, ra)
+		}
+	}
+	if len(res) == 0 {
+		return nil
+	}
+	return res
+}
+
 // Implement further consistency checks.
 func (snapdcl *SnapDeclaration) checkConsistency(db RODatabase, acck *AccountKey) error {
 	if !db.IsTrustedAccount(snapdcl.AuthorityID()) {
@@ -105,7 +125,7 @@ func (snapdcl *SnapDeclaration) checkConsistency(db RODatabase, acck *AccountKey
 	_, err := db.Find(AccountType, map[string]string{
 		"account-id": snapdcl.PublisherID(),
 	})
-	if IsNotFound(err) {
+	if errors.Is(err, &NotFoundError{}) {
 		return fmt.Errorf("snap-declaration assertion for %q (id %q) does not have a matching account assertion for the publisher %q", snapdcl.SnapName(), snapdcl.SnapID(), snapdcl.PublisherID())
 	}
 	if err != nil {
@@ -115,7 +135,7 @@ func (snapdcl *SnapDeclaration) checkConsistency(db RODatabase, acck *AccountKey
 	return nil
 }
 
-// sanity
+// expected interface is implemented
 var _ consistencyChecker = (*SnapDeclaration)(nil)
 
 // Prerequisites returns references to this snap-declaration's prerequisite assertions.
@@ -178,6 +198,9 @@ func snapDeclarationFormatAnalyze(headers map[string]interface{}, body []byte) (
 		if rule.feature(altAttrMatcherFeature) {
 			setFormatNum(5)
 		}
+		if rule.feature(publisherIDConstraintsFeature) {
+			setFormatNum(6)
+		}
 	})
 	if err != nil {
 		return 0, err
@@ -200,6 +223,9 @@ func snapDeclarationFormatAnalyze(headers map[string]interface{}, body []byte) (
 		if rule.feature(altAttrMatcherFeature) {
 			setFormatNum(5)
 		}
+		if rule.feature(publisherIDConstraintsFeature) {
+			setFormatNum(6)
+		}
 	})
 	if err != nil {
 		return 0, err
@@ -314,17 +340,182 @@ func assembleSnapDeclaration(assert assertionBase) (Assertion, error) {
 		return nil, err
 	}
 
+	var ras []*RevisionAuthority
+
+	ra, ok := assert.headers["revision-authority"]
+	if ok {
+		ramaps, ok := ra.([]interface{})
+		if !ok {
+			return nil, fmt.Errorf("revision-authority stanza must be a list of maps")
+		}
+		if len(ramaps) == 0 {
+			// there is no syntax producing this scenario but be robust
+			return nil, fmt.Errorf("revision-authority stanza cannot be empty")
+		}
+		ras = make([]*RevisionAuthority, 0, len(ramaps))
+		for _, ramap := range ramaps {
+			m, ok := ramap.(map[string]interface{})
+			if !ok {
+				return nil, fmt.Errorf("revision-authority stanza must be a list of maps")
+			}
+			accountID, err := checkStringMatchesWhat(m, "account-id", "in revision authority", validAccountID)
+			if err != nil {
+				return nil, err
+			}
+			prov, err := checkStringListInMap(m, "provenance", "provenance in revision authority", naming.ValidProvenance)
+			if err != nil {
+				return nil, err
+			}
+			if len(prov) == 0 {
+				return nil, fmt.Errorf("provenance in revision authority cannot be empty")
+			}
+			minRevision := 1
+			maxRevision := 0
+			if _, ok := m["min-revision"]; ok {
+				var err error
+				minRevision, err = checkSnapRevisionWhat(m, "min-revision", "in revision authority")
+				if err != nil {
+					return nil, err
+				}
+			}
+			if _, ok := m["max-revision"]; ok {
+				var err error
+				maxRevision, err = checkSnapRevisionWhat(m, "max-revision", "in revision authority")
+				if err != nil {
+					return nil, err
+				}
+			}
+			if maxRevision != 0 && maxRevision < minRevision {
+				return nil, fmt.Errorf("optional max-revision cannot be less than min-revision in revision-authority")
+			}
+			devscope, err := compileDeviceScopeConstraint(m, "revision-authority")
+			if err != nil {
+				return nil, err
+			}
+			ras = append(ras, &RevisionAuthority{
+				AccountID:   accountID,
+				Provenance:  prov,
+				MinRevision: minRevision,
+				MaxRevision: maxRevision,
+				DeviceScope: devscope,
+			})
+		}
+
+	}
+
 	return &SnapDeclaration{
-		assertionBase:  assert,
-		refreshControl: refControl,
-		plugRules:      plugRules,
-		slotRules:      slotRules,
-		autoAliases:    autoAliases,
-		aliases:        aliases,
-		timestamp:      timestamp,
+		assertionBase:       assert,
+		refreshControl:      refControl,
+		plugRules:           plugRules,
+		slotRules:           slotRules,
+		autoAliases:         autoAliases,
+		aliases:             aliases,
+		revisionAuthorities: ras,
+		timestamp:           timestamp,
 	}, nil
 }
 
+// RevisionAuthority holds information about an account that can sign revisions
+// for a given snap.
+type RevisionAuthority struct {
+	AccountID  string
+	Provenance []string
+
+	MinRevision int
+	MaxRevision int
+
+	DeviceScope *DeviceScopeConstraint
+}
+
+func (ra *RevisionAuthority) checkProvenanceAndRevision(a interface {
+	Assertion
+	Provenance() string
+}, what string, revno int, model *Model, store *Store) error {
+	if !strutil.ListContains(ra.Provenance, a.Provenance()) {
+		return fmt.Errorf("provenance mismatch")
+	}
+	if a.AuthorityID() != ra.AccountID {
+		return fmt.Errorf("authority-id mismatch")
+	}
+	if revno < ra.MinRevision {
+		return fmt.Errorf("%s revision %d is less than min-revision %d", what, revno, ra.MinRevision)
+	}
+	if ra.MaxRevision != 0 && revno > ra.MaxRevision {
+		return fmt.Errorf("%s revision %d is greater than max-revision %d", what, revno, ra.MaxRevision)
+	}
+	if ra.DeviceScope != nil && model != nil {
+		opts := DeviceScopeConstraintCheckOptions{UseFriendlyStores: true}
+		if err := ra.DeviceScope.Check(model, store, &opts); err != nil {
+			return err
+		}
+	}
+	return nil
+}
+
+// Check tests whether rev matches the revision authority constraints.
+// Optional model and store must be provided to cross-check device-specific
+// constraints.
+func (ra *RevisionAuthority) Check(rev *SnapRevision, model *Model, store *Store) error {
+	return ra.checkProvenanceAndRevision(rev, "snap", rev.SnapRevision(), model, store)
+}
+
+// CheckResourceRevision tests whether resrev matches the revision authority
+// constraints. Optional model and store must be provided to cross-check
+// device-specific constraints.
+func (ra *RevisionAuthority) CheckResourceRevision(resrev *SnapResourceRevision, model *Model, store *Store) error {
+	return ra.checkProvenanceAndRevision(resrev, "resource", resrev.ResourceRevision(), model, store)
+}
+
+var validSnapIntegrityTypes = []string{"dm-verity"}
+
+var validVersionsForIntegrityType = map[string][]int{
+	// version 1 corresponds to dm-verity format 1
+	"dm-verity": {1},
+}
+
+var validHashAlgorithmsForIntegrityType = map[string][]string{
+	// kernel supported algorithms:
+	// https://gitlab.com/cryptsetup/cryptsetup/-/blob/main/lib/crypto_backend/crypto_kernel.c?ref_type=heads#L35
+	// Go crypto's supported algorithms:
+	// https://cs.opensource.google/go/go/+/refs/tags/go1.23.4:src/crypto/crypto.go;l=68
+	"dm-verity": {
+		"sha256",
+	},
+}
+
+func contains[V int | string](l []V, i V) bool {
+	for _, v := range l {
+		if v == i {
+			return true
+		}
+	}
+
+	return false
+}
+
+func toHash(s string) crypto.Hash {
+	switch s {
+	case "sha256":
+		return crypto.SHA256
+	default:
+		return 0
+	}
+}
+
+// SnapIntegrityData holds information about integrity data of a specific type included in a snap's revision.
+//
+// A single snap revision can have multiple variants of integrity data which are represented as an array in the
+// snap revision assertion.
+type SnapIntegrityData struct {
+	Type          string
+	Version       uint
+	HashAlg       string
+	DataBlockSize uint
+	HashBlockSize uint
+	Digest        string
+	Salt          string
+}
+
 // SnapFileSHA3_384 computes the SHA3-384 digest of the given snap file.
 // It also returns its size.
 func SnapFileSHA3_384(snapPath string) (digest string, size uint64, err error) {
@@ -414,6 +605,8 @@ type SnapRevision struct {
 	snapSize     uint64
 	snapRevision int
 	timestamp    time.Time
+
+	snapIntegrityData []SnapIntegrityData
 }
 
 // SnapSHA3_384 returns the SHA3-384 digest of the snap.
@@ -421,6 +614,12 @@ func (snaprev *SnapRevision) SnapSHA3_384() string {
 	return snaprev.HeaderString("snap-sha3-384")
 }
 
+// Provenance returns the optional provenance of the snap (defaults to
+// global-upload (naming.DefaultProvenance)).
+func (snaprev *SnapRevision) Provenance() string {
+	return snaprev.HeaderString("provenance")
+}
+
 // SnapID returns the snap id of the snap.
 func (snaprev *SnapRevision) SnapID() string {
 	return snaprev.HeaderString("snap-id")
@@ -447,36 +646,59 @@ func (snaprev *SnapRevision) Timestamp() time.Time {
 	return snaprev.timestamp
 }
 
+// SnapIntegrityData returns the snap integrity data associated with the snap revision if any.
+func (snaprev *SnapRevision) SnapIntegrityData() []SnapIntegrityData {
+	return snaprev.snapIntegrityData
+}
+
 // Implement further consistency checks.
 func (snaprev *SnapRevision) checkConsistency(db RODatabase, acck *AccountKey) error {
-	// TODO: expand this to consider other stores signing on their own
-	if !db.IsTrustedAccount(snaprev.AuthorityID()) {
+	otherProvenance := snaprev.Provenance() != naming.DefaultProvenance
+	if !otherProvenance && !db.IsTrustedAccount(snaprev.AuthorityID()) {
+		// delegating global-upload revisions is not allowed
 		return fmt.Errorf("snap-revision assertion for snap id %q is not signed by a store: %s", snaprev.SnapID(), snaprev.AuthorityID())
 	}
 	_, err := db.Find(AccountType, map[string]string{
 		"account-id": snaprev.DeveloperID(),
 	})
-	if IsNotFound(err) {
+	if errors.Is(err, &NotFoundError{}) {
 		return fmt.Errorf("snap-revision assertion for snap id %q does not have a matching account assertion for the developer %q", snaprev.SnapID(), snaprev.DeveloperID())
 	}
 	if err != nil {
 		return err
 	}
-	_, err = db.Find(SnapDeclarationType, map[string]string{
+	a, err := db.Find(SnapDeclarationType, map[string]string{
 		// XXX: mediate getting current series through some context object? this gets the job done for now
 		"series":  release.Series,
 		"snap-id": snaprev.SnapID(),
 	})
-	if IsNotFound(err) {
+	if errors.Is(err, &NotFoundError{}) {
 		return fmt.Errorf("snap-revision assertion for snap id %q does not have a matching snap-declaration assertion", snaprev.SnapID())
 	}
 	if err != nil {
 		return err
 	}
+	if otherProvenance {
+		decl := a.(*SnapDeclaration)
+		ras := decl.RevisionAuthority(snaprev.Provenance())
+		matchingRevAuthority := false
+		for _, ra := range ras {
+			// model==store==nil, we do not perform device-specific
+			// checks at this level, those are performed at
+			// higher-level guarding installing actual snaps
+			if err := ra.Check(snaprev, nil, nil); err == nil {
+				matchingRevAuthority = true
+				break
+			}
+		}
+		if !matchingRevAuthority {
+			return fmt.Errorf("snap-revision assertion with provenance %q for snap id %q is not signed by an authorized authority: %s", snaprev.Provenance(), snaprev.SnapID(), snaprev.AuthorityID())
+		}
+	}
 	return nil
 }
 
-// sanity
+// expected interface is implemented
 var _ consistencyChecker = (*SnapRevision)(nil)
 
 // Prerequisites returns references to this snap-revision's prerequisite assertions.
@@ -499,12 +721,114 @@ func checkSnapRevisionWhat(headers map[string]interface{}, name, what string) (s
 	return snapRevision, nil
 }
 
+func checkOptionalSnapRevisionWhat(headers map[string]interface{}, name, what string) (snapRevision int, err error) {
+	if _, ok := headers[name]; !ok {
+		return 0, nil
+	}
+	return checkSnapRevisionWhat(headers, name, what)
+}
+
+func checkSnapIntegrity(headers map[string]interface{}) ([]SnapIntegrityData, error) {
+	value, ok := headers["integrity"]
+	if !ok {
+		// integrity stanzas are optional
+		return nil, nil
+	}
+
+	integrityList, ok := value.([]interface{})
+	if !ok {
+		return nil, fmt.Errorf(`"integrity" header must contain a list of integrity data`)
+	}
+	if len(integrityList) == 0 {
+		return nil, nil
+	}
+
+	var snapIntegrityDataList []SnapIntegrityData
+
+	for i, il := range integrityList {
+		id, ok := il.(map[string]interface{})
+		if !ok {
+			return nil, fmt.Errorf(`"integrity" header must contain a list of integrity data`)
+		}
+
+		what := fmt.Sprintf("of integrity data [%d]", i)
+		typ, err := checkExistsStringWhat(id, "type", what)
+		if err != nil {
+			return nil, err
+		}
+
+		if !contains(validSnapIntegrityTypes, typ) {
+			return nil, fmt.Errorf("\"type\" of integrity data [%d] must be one of (%s)", i, strings.Join(validSnapIntegrityTypes, "|"))
+		}
+
+		what = fmt.Sprintf("of integrity data [%d] of type %q", i, typ)
+		version, err := checkUintWhat(id, "version", 64, what)
+		if err != nil {
+			return nil, err
+		}
+
+		if !contains(validVersionsForIntegrityType[typ], int(version)) {
+			return nil, fmt.Errorf(`version of integrity data [%d] of type %q must be one of %v`, i, typ, validVersionsForIntegrityType[typ])
+		}
+
+		alg, err := checkExistsStringWhat(id, "hash-algorithm", what)
+		if err != nil {
+			return nil, err
+		}
+
+		if !contains(validHashAlgorithmsForIntegrityType[typ], alg) {
+			return nil, fmt.Errorf(`hash algorithm of integrity data [%d] of type %q must be one of %v`, i, typ, validHashAlgorithmsForIntegrityType[typ])
+		}
+
+		what = fmt.Sprintf("of integrity data [%d] of type %q (%s)", i, typ, alg)
+		dataBlockSize, err := checkUintWhat(id, "data-block-size", 64, what)
+		if err != nil {
+			return nil, err
+		}
+
+		hashBlockSize, err := checkUintWhat(id, "hash-block-size", 64, what)
+		if err != nil {
+			return nil, err
+		}
+
+		h := toHash(alg)
+		encDigest, err := checkDigestDecWhat(id, "digest", h, hex.DecodeString, what)
+		if err != nil {
+			return nil, err
+		}
+
+		encSalt, err := checkDigestDecWhat(id, "salt", h, hex.DecodeString, what)
+		if err != nil {
+			return nil, err
+		}
+
+		snapIntegrityData := SnapIntegrityData{
+			Type:          typ,
+			Version:       uint(version),
+			HashAlg:       alg,
+			DataBlockSize: uint(dataBlockSize),
+			HashBlockSize: uint(hashBlockSize),
+			Digest:        encDigest,
+			Salt:          encSalt,
+		}
+
+		snapIntegrityDataList = append(snapIntegrityDataList, snapIntegrityData)
+	}
+
+	return snapIntegrityDataList, nil
+}
+
 func assembleSnapRevision(assert assertionBase) (Assertion, error) {
 	_, err := checkDigest(assert.headers, "snap-sha3-384", crypto.SHA3_384)
 	if err != nil {
 		return nil, err
 	}
 
+	_, err = checkStringMatches(assert.headers, "provenance", naming.ValidProvenance)
+	if err != nil {
+		return nil, err
+	}
+
 	_, err = checkNotEmptyString(assert.headers, "snap-id")
 	if err != nil {
 		return nil, err
@@ -530,11 +854,17 @@ func assembleSnapRevision(assert assertionBase) (Assertion, error) {
 		return nil, err
 	}
 
+	snapIntegrityData, err := checkSnapIntegrity(assert.headers)
+	if err != nil {
+		return nil, err
+	}
+
 	return &SnapRevision{
-		assertionBase: assert,
-		snapSize:      snapSize,
-		snapRevision:  snapRevision,
-		timestamp:     timestamp,
+		assertionBase:     assert,
+		snapSize:          snapSize,
+		snapRevision:      snapRevision,
+		timestamp:         timestamp,
+		snapIntegrityData: snapIntegrityData,
 	}, nil
 }
 
@@ -585,7 +915,7 @@ func (validation *Validation) checkConsistency(db RODatabase, acck *AccountKey)
 		"series":  validation.Series(),
 		"snap-id": validation.ApprovedSnapID(),
 	})
-	if IsNotFound(err) {
+	if errors.Is(err, &NotFoundError{}) {
 		return fmt.Errorf("validation assertion by snap-id %q does not have a matching snap-declaration assertion for approved-snap-id %q", validation.SnapID(), validation.ApprovedSnapID())
 	}
 	if err != nil {
@@ -595,7 +925,7 @@ func (validation *Validation) checkConsistency(db RODatabase, acck *AccountKey)
 		"series":  validation.Series(),
 		"snap-id": validation.SnapID(),
 	})
-	if IsNotFound(err) {
+	if errors.Is(err, &NotFoundError{}) {
 		return fmt.Errorf("validation assertion by snap-id %q does not have a matching snap-declaration assertion", validation.SnapID())
 	}
 	if err != nil {
@@ -610,7 +940,7 @@ func (validation *Validation) checkConsistency(db RODatabase, acck *AccountKey)
 	return nil
 }
 
-// sanity
+// expected interface is implemented
 var _ consistencyChecker = (*Validation)(nil)
 
 // Prerequisites returns references to this validation's prerequisite assertions.
@@ -684,7 +1014,7 @@ func (basedcl *BaseDeclaration) checkConsistency(db RODatabase, acck *AccountKey
 	return nil
 }
 
-// sanity
+// expected interface is implemented
 var _ consistencyChecker = (*BaseDeclaration)(nil)
 
 func assembleBaseDeclaration(assert assertionBase) (Assertion, error) {
@@ -828,7 +1158,7 @@ func (snapdev *SnapDeveloper) checkConsistency(db RODatabase, acck *AccountKey)
 		"snap-id": snapdev.SnapID(),
 	})
 	if err != nil {
-		if IsNotFound(err) {
+		if errors.Is(err, &NotFoundError{}) {
 			return fmt.Errorf("snap-developer assertion for snap id %q does not have a matching snap-declaration assertion", snapdev.SnapID())
 		}
 		return err
@@ -837,7 +1167,7 @@ func (snapdev *SnapDeveloper) checkConsistency(db RODatabase, acck *AccountKey)
 	// check there's an account for the publisher-id
 	_, err = db.Find(AccountType, map[string]string{"account-id": publisherID})
 	if err != nil {
-		if IsNotFound(err) {
+		if errors.Is(err, &NotFoundError{}) {
 			return fmt.Errorf("snap-developer assertion for snap-id %q does not have a matching account assertion for the publisher %q", snapdev.SnapID(), publisherID)
 		}
 		return err
@@ -850,7 +1180,7 @@ func (snapdev *SnapDeveloper) checkConsistency(db RODatabase, acck *AccountKey)
 		}
 		_, err = db.Find(AccountType, map[string]string{"account-id": developerID})
 		if err != nil {
-			if IsNotFound(err) {
+			if errors.Is(err, &NotFoundError{}) {
 				return fmt.Errorf("snap-developer assertion for snap-id %q does not have a matching account assertion for the developer %q", snapdev.SnapID(), developerID)
 			}
 			return err
@@ -860,7 +1190,7 @@ func (snapdev *SnapDeveloper) checkConsistency(db RODatabase, acck *AccountKey)
 	return nil
 }
 
-// sanity
+// expected interface is implemented
 var _ consistencyChecker = (*SnapDeveloper)(nil)
 
 // Prerequisites returns references to this snap-developer's prerequisite assertions.
