@@ -88,6 +88,7 @@ func sessionInfo(c *Command, r *http.Request) Response {
 	}
 	return SyncResponse(m)
 }
+
 func serviceStart(inst *client.ServiceInstruction, sysd systemd.Systemd) Response {
 	// Refuse to start non-snap services
 	for _, service := range inst.Services {
@@ -97,6 +98,32 @@ func serviceStart(inst *client.ServiceInstruction, sysd systemd.Systemd) Respons
 	}
 
 	startErrors := make(map[string]string)
+	var err error
+	if inst.Enable {
+		if err = sysd.EnableNoReload(inst.Services); err != nil {
+			return InternalError("cannot enable snap services %q: %v", inst.Services, err)
+		}
+
+		// Setup undo logic for the enable in case of errors
+		defer func() {
+			if err == nil && len(startErrors) == 0 {
+				return
+			}
+
+			// Only log errors in this case to avoid overriding the initial error
+			if err := sysd.DisableNoReload(inst.Services); err != nil {
+				logger.Noticef("cannot disable previously enabled services %q: %v", inst.Services, err)
+			}
+			if err := sysd.DaemonReload(); err != nil {
+				logger.Noticef("cannot reload systemd: %v", err)
+			}
+		}()
+
+		if err = sysd.DaemonReload(); err != nil {
+			return InternalError("cannot reload systemd: %v", err)
+		}
+	}
+
 	var started []string
 	for _, service := range inst.Services {
 		if err := sysd.Start([]string{service}); err != nil {
@@ -105,18 +132,20 @@ func serviceStart(inst *client.ServiceInstruction, sysd systemd.Systemd) Respons
 		}
 		started = append(started, service)
 	}
-	// If we got any failures, attempt to stop the services we started.
-	stopErrors := make(map[string]string)
-	if len(startErrors) != 0 {
-		for _, service := range started {
-			if err := sysd.Stop([]string{service}); err != nil {
-				stopErrors[service] = err.Error()
-			}
-		}
-	}
+
 	if len(startErrors) == 0 {
 		return SyncResponse(nil)
 	}
+
+	// If we got any failures, attempt to stop the services we started, and
+	// then re-disable if enable was requested
+	stopErrors := make(map[string]string)
+	for _, service := range started {
+		if err := sysd.Stop([]string{service}); err != nil {
+			stopErrors[service] = err.Error()
+		}
+	}
+
 	return SyncResponse(&resp{
 		Type:   ResponseTypeError,
 		Status: 500,
@@ -181,20 +210,30 @@ func serviceStop(inst *client.ServiceInstruction, sysd systemd.Systemd) Response
 			stopErrors[service] = err.Error()
 		}
 	}
-	if len(stopErrors) == 0 {
-		return SyncResponse(nil)
-	}
-	return SyncResponse(&resp{
-		Type:   ResponseTypeError,
-		Status: 500,
-		Result: &errorResult{
-			Message: "some user services failed to stop",
-			Kind:    errorKindServiceControl,
-			Value: map[string]interface{}{
-				"stop-errors": stopErrors,
+
+	if len(stopErrors) != 0 {
+		return SyncResponse(&resp{
+			Type:   ResponseTypeError,
+			Status: 500,
+			Result: &errorResult{
+				Message: "some user services failed to stop",
+				Kind:    errorKindServiceControl,
+				Value: map[string]interface{}{
+					"stop-errors": stopErrors,
+				},
 			},
-		},
-	})
+		})
+	}
+
+	if inst.Disable {
+		if err := sysd.DisableNoReload(inst.Services); err != nil {
+			return InternalError(fmt.Sprintf("cannot disable services %q: %v", inst.Services, err))
+		}
+		if err := sysd.DaemonReload(); err != nil {
+			return InternalError(fmt.Sprintf("cannot reload systemd: %v", err))
+		}
+	}
+	return SyncResponse(nil)
 }
 
 func serviceDaemonReload(inst *client.ServiceInstruction, sysd systemd.Systemd) Response {
@@ -319,6 +358,29 @@ func serviceStatus(c *Command, r *http.Request) Response {
 	return SyncResponse(unitStatusToClientUnitStatus(stss))
 }
 
+var currentLocale = i18n.CurrentLocale
+
+func getLocalizedAppNameFromDesktopFile(parser *goconfigparser.ConfigParser, defaultName string) string {
+	// First try with full locale string (e.g. es_ES)
+	locale := fmt.Sprintf("Name[%s]", currentLocale())
+	if name, err := parser.Get("Desktop Entry", locale); err == nil && name != "" {
+		return name
+	}
+
+	// If not found, try with the country part
+	locale = fmt.Sprintf("Name[%s]", strings.Split(currentLocale(), "_")[0])
+	if name, err := parser.Get("Desktop Entry", locale); err == nil && name != "" {
+		return name
+	}
+
+	// If neither are found, try with the untranslated name
+	if name, err := parser.Get("Desktop Entry", "Name"); err == nil && name != "" {
+		return name
+	}
+
+	return defaultName
+}
+
 func postPendingRefreshNotification(c *Command, r *http.Request) Response {
 	if ok, resp := validateJSONRequest(r); !ok {
 		return resp
@@ -344,11 +406,26 @@ func postPendingRefreshNotification(c *Command, r *http.Request) Response {
 	}
 
 	// TODO: this message needs to be crafted better as it's the only thing guaranteed to be delivered.
-	summary := fmt.Sprintf(i18n.G("Update available for %s."), refreshInfo.InstanceName)
 	var urgencyLevel notification.Urgency
-	var body, icon string
+	var body, icon, name string
 	var hints []notification.Hint
 
+	snapname, instanceKey := snap.SplitInstanceName(refreshInfo.InstanceName)
+	// If we have a desktop file of the busy application, use that apps's icon and name, if possible
+	if refreshInfo.BusyAppDesktopEntry != "" {
+		parser := goconfigparser.New()
+		desktopFilePath := filepath.Join(dirs.SnapDesktopFilesDir, refreshInfo.BusyAppDesktopEntry+".desktop")
+		if err := parser.ReadFile(desktopFilePath); err == nil {
+			icon, _ = parser.Get("Desktop Entry", "Icon")
+			name = combineNameAndKey(getLocalizedAppNameFromDesktopFile(parser, snapname), instanceKey)
+		}
+	}
+	if name == "" {
+		name = combineNameAndKey(snapname, instanceKey)
+	}
+
+	summary := fmt.Sprintf(i18n.G("Update available for %s."), name)
+
 	if daysLeft := int(refreshInfo.TimeRemaining.Truncate(time.Hour).Hours() / 24); daysLeft > 0 {
 		urgencyLevel = notification.LowUrgency
 		body = fmt.Sprintf(
@@ -365,20 +442,12 @@ func postPendingRefreshNotification(c *Command, r *http.Request) Response {
 			i18n.NG("Close the application to update now. It will update automatically in %d minute.",
 				"Close the application to update now. It will update automatically in %d minutes.", minutesLeft), minutesLeft)
 	} else {
-		summary = fmt.Sprintf(i18n.G("%s is updating now!"), refreshInfo.InstanceName)
+		summary = fmt.Sprintf(i18n.G("%s is updating now!"), name)
 		urgencyLevel = notification.CriticalUrgency
 	}
 	hints = append(hints, notification.WithUrgency(urgencyLevel))
 	// The notification is provided by snapd session agent.
 	hints = append(hints, notification.WithDesktopEntry("io.snapcraft.SessionAgent"))
-	// But if we have a desktop file of the busy application, use that apps's icon.
-	if refreshInfo.BusyAppDesktopEntry != "" {
-		parser := goconfigparser.New()
-		desktopFilePath := filepath.Join(dirs.SnapDesktopFilesDir, refreshInfo.BusyAppDesktopEntry+".desktop")
-		if err := parser.ReadFile(desktopFilePath); err == nil {
-			icon, _ = parser.Get("Desktop Entry", "Icon")
-		}
-	}
 
 	msg := &notification.Message{
 		AppName: refreshInfo.BusyAppName,
@@ -402,8 +471,7 @@ func postPendingRefreshNotification(c *Command, r *http.Request) Response {
 	return SyncResponse(nil)
 }
 
-func guessAppIcon(si *snap.Info) string {
-	var icon string
+func guessAppData(si *snap.Info, defaultName string, instanceKey string) (icon string, name string) {
 	parser := goconfigparser.New()
 
 	// trivial heuristic, if the app is named like a snap then
@@ -413,11 +481,13 @@ func guessAppIcon(si *snap.Info) string {
 	if ok && !mainApp.IsService() {
 		// got the main app, grab its desktop file
 		if err := parser.ReadFile(mainApp.DesktopFile()); err == nil {
+			name = combineNameAndKey(getLocalizedAppNameFromDesktopFile(parser, defaultName), instanceKey)
 			icon, _ = parser.Get("Desktop Entry", "Icon")
 		}
 	}
+
 	if icon != "" {
-		return icon
+		return icon, name
 	}
 
 	// If it doesn't exist, take the first app in the snap with a DesktopFile with icon
@@ -426,12 +496,22 @@ func guessAppIcon(si *snap.Info) string {
 			continue
 		}
 		if err := parser.ReadFile(app.DesktopFile()); err == nil {
+			name = combineNameAndKey(getLocalizedAppNameFromDesktopFile(parser, defaultName), instanceKey)
 			if icon, err = parser.Get("Desktop Entry", "Icon"); err == nil && icon != "" {
 				break
 			}
 		}
 	}
-	return icon
+
+	return icon, name
+}
+
+func combineNameAndKey(name, key string) string {
+	if key != "" {
+		return fmt.Sprintf("%s (%s)", name, key)
+	} else {
+		return name
+	}
 }
 
 func postRefreshFinishedNotification(c *Command, r *http.Request) Response {
@@ -446,6 +526,17 @@ func postRefreshFinishedNotification(c *Command, r *http.Request) Response {
 		return BadRequest("cannot decode request body into finish refresh notification info: %v", err)
 	}
 
+	var icon string
+	name, instanceKey := snap.SplitInstanceName(finishRefresh.InstanceName)
+	if si, err := snap.ReadCurrentInfo(finishRefresh.InstanceName); err == nil {
+		icon, name = guessAppData(si, name, instanceKey)
+	} else {
+		logger.Noticef("cannot load snap-info for %s: %v", combineNameAndKey(name, instanceKey), err)
+	}
+	if name == "" {
+		name = combineNameAndKey(name, instanceKey)
+	}
+
 	// Note that since the connection is shared, we are not closing it.
 	if c.s.bus == nil {
 		return SyncResponse(&resp{
@@ -457,27 +548,20 @@ func postRefreshFinishedNotification(c *Command, r *http.Request) Response {
 		})
 	}
 
-	summary := fmt.Sprintf(i18n.G("%s was updated."), finishRefresh.InstanceName)
+	summary := fmt.Sprintf(i18n.G("%s was updated."), name)
 	body := i18n.G("Ready to launch.")
 	hints := []notification.Hint{
 		notification.WithDesktopEntry("io.snapcraft.SessionAgent"),
 		notification.WithUrgency(notification.LowUrgency),
 	}
 
-	var icon string
-	if si, err := snap.ReadCurrentInfo(finishRefresh.InstanceName); err == nil {
-		icon = guessAppIcon(si)
-	} else {
-		logger.Noticef("cannot load snap-info for %s: %v", finishRefresh.InstanceName, err)
-	}
-
 	msg := &notification.Message{
 		Title: summary,
 		Body:  body,
 		Hints: hints,
 		Icon:  icon,
 	}
-	if err := c.s.notificationMgr.SendNotification(notification.ID(finishRefresh.InstanceName), msg); err != nil {
+	if err := c.s.notificationMgr.SendNotification(notification.ID(name), msg); err != nil {
 		return SyncResponse(&resp{
 			Type:   ResponseTypeError,
 			Status: 500,
